-- atlas.lua
-- Original code by Santos.
-- Any licensing and copyright to be determine by him.
-- Code modified by Ref for own use 07.17.2013
-- =================================================================================
-- Functions:
--		atlas.make					-- combines images (in a directory) into a spritesheet
-- 			requires:
--				directory name		-- source of images
--			returns (when 's' pressed) to AppData directory:
--				spritesheet#.png
--				geometry#.txt
--		atlas.load					-- retreaves images from atlas
-- 			requires:
--				spritesheet#.png
--				geometry#.txt
--			returns:
--				atlas.geometries	-- quad for each image
--				atlas.names			-- name of each image (provide number => get name)
--				atlas.num			-- quad number of image (provide name => get number)
--				atlas.size			-- quad size {x,y} for centering image - So sad!
--			to disply image:
--				gr.drawq( at.sheet, at.geometries[#],x,y,angle,sizeX,sizeY,at.size[#][1]/2,at.size[#][2])
-- ===================================================================================

local atlas = {names={}, images={}, geometries = {}, data = {}, num = {}, size = {} }
function atlas.make( directory )	-- id = image data, pid = padded image data
	for _, filename in pairs(love.filesystem.enumerate( directory )) do	-- get files in directory
		if love.filesystem.isFile(directory..'/'..filename) then
			local name, ext = filename:match('(.+)%.(.+)$')
			ext = ext:lower()	-- only use image files
			if ext == 'png' or ext == 'bmp' or ext == 'jpg' or ext == 'jpeg' or ext == 'gif' or ext == 'tga' then
				local id = love.image.newImageData(directory..'/'..filename)
				-- ***** add one pixel data point around each image (to prevent edge 'bleed') *****
				local pid = love.image.newImageData(id:getWidth() + 2, id:getHeight() + 2)
				pid:paste( id, 1, 1, 0, 0, id:getWidth(), id:getHeight())
				pid:paste( id, 0, 1, 0, 0, 1, id:getHeight())
				pid:paste( id, pid:getWidth()-1, 1, id:getWidth()-1, 0, 1, id:getHeight())
				pid:paste( id, 1, 0, 0, 0, id:getWidth(), 1)
				pid:paste( id, 1, pid:getHeight()-1, 0, id:getHeight()-1, id:getWidth(), 1)
				pid:setPixel( 0, 0, id:getPixel(0, 0))
				pid:setPixel( pid:getWidth()-1, 0, id:getPixel(id:getWidth()-1, 0))
				pid:setPixel( pid:getWidth()-1, pid:getHeight()-1, id:getPixel(id:getWidth()-1, id:getHeight()-1))
				pid:setPixel( 0, pid:getHeight()-1, id:getPixel(0, id:getHeight()-1))
				-- ***********************************************************************
				table.insert(atlas.images, { name = name, imagedata = pid, width = pid:getWidth(),
										height = pid:getHeight() })
				atlas.names[#atlas.names+1] = name
				end
			end
		end					-- note: table.sort will place name2 after name11
	table.sort( atlas.images, function(a, b) return a.height > b.height end )
	local total, max_width = 0, 0
	for i, image in ipairs( atlas.images ) do
		total = total + image.width * image.height
		if image.width > max_width then max_width = image.width end
		end
	-- ********** atlas width ****************** --
	local estimated_atlas_width
	if max_width > math.sqrt( total ) then
		estimated_atlas_width = max_width
		else
		estimated_atlas_width = math.ceil( math.sqrt( total ) )
		end
	local imageY, imageX, width_of_atlas = 0, 0, 0
	local first_height	 = atlas.images[1] and atlas.images[1].height
	for i, image in ipairs(atlas.images) do
	if ( imageX + image.width ) > estimated_atlas_width then
		imageX = 0
		imageY = imageY + first_height
		first_height = image.height
		end
	atlas.images[i].x, atlas.images[i].y = imageX, imageY
	imageX  = imageX + image.width
	if imageX > width_of_atlas then width_of_atlas = imageX end
	end
	-- *********** atlas height *********************** --
	local height_of_atlas = atlas.images[#atlas.images].y + first_height
	-- ************************************************ --
	atlas.sheet = love.image.newImageData( width_of_atlas, height_of_atlas )
	love.graphics.setBackgroundColor( 255, 255, 255, 0 )	-- needed for transparent images
	local wd, ht = atlas.sheet:getWidth(), atlas.sheet:getHeight()
	love.graphics.setMode( wd, ht, {} )
	for i, image in ipairs( atlas.images ) do
		atlas.sheet:paste( image.imagedata, image.x, image.y, 0, 0, image.width, image.height )
		atlas.geometries[image.name] = love.graphics.newQuad(image.x+1, image.y+1, image.width-2, image.height-2, wd, ht)
		atlas.data[i] = atlas.names[i]..','..(image.x+1)..","..(image.y+1)..','..(image.width-2)..','..(image.height-2)..','..wd..','..ht..'\n\r'
		end
	atlas.sheet = love.graphics.newImage( atlas.sheet )	-- convert imagedata to image
	end

function atlas.storeData( name, dat, mode )	-- mode 'w' replaces, 'a' appends
	local bad_name	= true
	local num		= 1
	while bad_name do
		if love.filesystem.exists(name..num..'.txt'  ) then
			num = num + 1
			else
			bad_name = false
			end
		end
	local file = love.filesystem.newFile( name..num..'.txt' )
	local mode = mode or 'w'
	file:open( mode )
	for i = 1, #dat do
		success = file:write( dat[i].."\n" )
		end
	file:close()
	end

function atlas.save_image(filename)	-- won't overwrite pre-existing files
	local bad_name	= true
	local num		= 1
	while bad_name do
		if love.filesystem.exists(filename..num..'.png'  ) then
			num = num + 1
			else
			bad_name = false
			end
		end
	local image	= love.graphics.newScreenshot()
	image:encode( filename..num..'.png' )
	end

function atlas:update()
	if save then
		save = false
		atlas.save_image( 'spritesheet' )
		atlas.storeData( 'geometry', atlas.data )
		end
	end

function atlas.keypressed( key )
	if key == 's' then save = true end
	end

function string:split( sep )
	local sep, fields = sep or ",", {}
	local pattern = string.format( '([^%s]+)', sep )
	self:gsub( pattern, function( c ) fields[#fields+1] = c end )
	return fields
	end

function atlas.load( sheet, geometry ) -- Returns: quads, quad names, quad sizes, index of quad name (name => number)
	atlas.sheet = love.graphics.newImage( sheet )
	local tab = {}
	for line in love.filesystem.lines( geometry ) do
		if #line>0 then tab[#tab+1]=line:split() end
		end
	for j = 1, #tab do						-- separate image names from geometry
		local temp = {}
		atlas.names[#atlas.names+1] = tab[j][1]
		for i = 2, #tab[j] do temp[i-1] = tab[j][i] end
		atlas.geometries[j] = love.graphics.newQuad( unpack( temp ) )
		atlas.size[j] = {temp[3],temp[4]}	-- saves image sizes ( for centering in draw )
		end									-- Oh so sad, inverted 'y' & non-centered images
	for i = 1, #atlas.names do atlas.num[atlas.names[i]] = i end	-- permits calling geometries by name
	end

return atlas
